#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2014-2015 Jérémy Bobbio <lunar@debian.org>
# Copyright © 2016-2017, 2020, 2023 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

from .tools import get_tool_name, get_package_provider


class OutputParsingError(Exception):
    def __init__(self, operation, object):
        super().__init__()
        self.operation = operation
        self.object_class = object.__class__


class RequiredToolNotFound(Exception):
    def __init__(self, operation):
        super().__init__()
        self.operation = get_tool_name(operation)

    def get_comment(self, infix=""):
        xs = [
            "'{}' not available in path.".format(self.operation),
            infix,
        ]

        x = get_package_provider(self.operation)
        if x:
            xs.append(
                "Installing the '{}' package may produce better output.".format(
                    x
                )
            )

        return " ".join(x for x in xs if x)


class ContainerExtractionError(Exception):
    def __init__(self, pathname, wrapped_exc):
        super().__init__()
        self.pathname = pathname
        self.wrapped_exc = wrapped_exc



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2014-2015 Jérémy Bobbio <lunar@debian.org>
# Copyright © 2017, 2019-2020 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import re
import hashlib
import logging
import subprocess
import functools

from .config import Config
from .profiling import profile

logger = logging.getLogger(__name__)

DIFF_CHUNK = 4096


@functools.lru_cache(maxsize=128)
def compile_string_regex(regex_str):
    return re.compile(regex_str)


@functools.lru_cache(maxsize=128)
def compile_bytes_regex(regex_str):
    return re.compile(regex_str.encode("utf-8"))


def filter_reader(buf, additional_filter=None):
    # Apply the passed filter first, for example Command.filter
    if additional_filter:
        buf = additional_filter(buf)

    # No need to work on empty lines
    if not buf:
        return buf

    # Use either str or bytes objects depending on buffer type
    if isinstance(buf, str):
        compile_func = compile_string_regex
        replace = "[masked]"
    else:
        compile_func = compile_bytes_regex
        replace = b"[masked]"

    for regex in Config().diff_masks:
        buf = compile_func(regex).sub(replace, buf)

    return buf


def from_raw_reader(in_file, filter=None):
    def feeder(out_file):
        max_lines = Config().max_diff_input_lines
        end_nl = False
        line_count = 0

        # If we have a maximum size, hash the content as we go along so we can
        # display a nicer message.
        h = None
        if max_lines < float("inf"):
            h = hashlib.sha256()

        for buf in in_file:
            line_count += 1
            out = filter_reader(buf, filter)

            if h is not None:
                h.update(out)

            if line_count < max_lines:
                out_file.write(out)
                # very long lines can sometimes interact negatively with
                # python buffering; force a flush here to avoid this,
                # see https://bugs.debian.org/870049
                out_file.flush()
            if buf:
                end_nl = buf[-1] == "\n"

        if h is not None and line_count >= max_lines:
            out_file.write(
                "[ Too much input for diff (SHA256: {}) ]\n".format(
                    h.hexdigest()
                ).encode("utf-8")
            )
            end_nl = True

        return end_nl

    return feeder


def from_text_reader(in_file, filter=None):
    if filter is None:

        def encoding_filter(text_buf):
            return text_buf.encode("utf-8")

    else:

        def encoding_filter(text_buf):
            return filter(text_buf).encode("utf-8")

    return from_raw_reader(in_file, encoding_filter)


def from_operation(operation):
    def feeder(out_file):
        with profile("command", operation.name):
            feeder = from_raw_reader(operation.output, operation.filter)
            end_nl = feeder(out_file)

        if operation.should_show_error():
            # On error, default to displaying all lines of the error
            output = operation.error_string or ""
            if not output and operation.output:
                # ... but if we don't have, return the first line of the
                # standard output.
                output = "{}{}".format(
                    operation.output[0].decode("utf-8", "ignore").strip(),
                    "\n[…]" if len(operation.output) > 1 else "",
                )
            raise subprocess.CalledProcessError(
                operation.returncode,
                [operation.full_name()],
                output=output.encode("utf-8"),
            )
        return end_nl

    return feeder


def from_text(content):
    """
    Works for both bytes and str objects.
    """

    def feeder(f):
        for offset in range(0, len(content), DIFF_CHUNK):
            buf = filter_reader(content[offset : offset + DIFF_CHUNK])
            if isinstance(buf, str):
                f.write(buf.encode("utf-8"))
            else:
                f.write(buf)

        return content and content[-1] == "\n"

    return feeder


def empty():
    def feeder(f):
        return False

    return feeder



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2017-2020, 2022-2023 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import os
import re
import glob
import logging

from diffoscope.tools import (
    python_module_missing,
    get_comment_for_missing_python_module,
)
from diffoscope.tempfiles import get_temporary_directory

from .utils.file import File
from .utils.archive import Archive

try:
    from .rpm import RpmFile
except ImportError:
    from .rpm_fallback import RpmFile

try:
    import binwalk
except Exception:
    python_module_missing("binwalk")
    binwalk = None
else:
    # Disable binwalk's own user configuration for predictable results and to
    # ensure it does not create (!) unnecessary directories, etc. (re. #903444)
    def fn(self):
        if not hasattr(fn, "_temp_dir"):
            fn._temp_dir = get_temporary_directory("binwalk")
        return fn._temp_dir.name

    binwalk.core.settings.Settings._get_user_config_dir = fn

logger = logging.getLogger(__name__)


class BinwalkFileContainer(Archive):
    def open_archive(self):
        return self

    def close_archive(self):
        self.source._unpacked.cleanup()

    def get_member_names(self):
        return sorted(self.source._members.keys())

    def extract(self, member_name, dest_dir):
        return self.source._members[member_name]


class BinwalkFile(File):
    FILE_TYPE_RE = re.compile(r"\bcpio archive\b")
    CONTAINER_CLASSES = [BinwalkFileContainer]

    @classmethod
    def recognizes(cls, file):
        if not super().recognizes(file):
            return False

        if binwalk is None:
            file.add_comment(get_comment_for_missing_python_module("binwalk"))
            return False

        # RPM files are .cpio, but let's always leave it to the RPM comparator.
        # This avoids the RpmFile > BinWalkFile > CpioFile > RpmFile circular
        # ordering dependency.
        if file.container and isinstance(file.container.source, RpmFile):
            return False

        # Don't recurse; binwalk has already found everything
        if isinstance(file.container, cls.CONTAINER_CLASSES[0]):
            return False

        unpacked = get_temporary_directory(prefix="binwalk")
        logger.debug("Extracting %s to %s", file.path, unpacked.name)

        try:
            binwalk.scan(
                file.path,
                dd="cpio:cpio",
                carve=True,
                quiet=True,
                signature=True,
                directory=unpacked.name,
            )
        except binwalk.core.exceptions.ModuleException:
            return False

        members = {
            "{} file embedded at offset {}".format(
                os.path.splitext(x)[1],
                os.path.basename(os.path.splitext(x)[0]),
            ): x
            for x in glob.glob(os.path.join(unpacked.name, "*/*"))
        }

        logger.debug("Found %d embedded member(s)", len(members))

        if not members:
            unpacked.cleanup()
            return False

        file._members = members
        file._unpacked = unpacked

        return True

    def compare(self, other, source=None):
        difference = super().compare(other, source)

        if difference is not None:
            difference.add_comment(
                "comprises of {} embedded members".format(len(self._members))
            )

        return difference



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2015 Reiner Herrmann <reiner@reiner-h.de>
# Copyright © 2015 Jérémy Bobbio <lunar@debian.org>
# Copyright © 2015-2021 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import os
import re
import stat
import logging
import functools
import subprocess
import collections

from diffoscope.tools import tool_required
from diffoscope.difference import Difference
from diffoscope.tempfiles import get_temporary_directory

from .utils.file import File
from .device import Device
from .symlink import Symlink
from .socket_or_fifo import SocketOrFIFO
from .directory import Directory
from .utils.archive import Archive, ArchiveMember
from .utils.command import Command

logger = logging.getLogger(__name__)


class SquashfsSuperblock(Command):
    @tool_required("unsquashfs")
    def cmdline(self):
        return ["unsquashfs", "-s", self.path]

    def filter(self, line):
        # strip filename
        return re.sub(
            r"^(Found a valid .*) on .*", "\\1", line.decode("utf-8")
        ).encode("utf-8")


class SquashfsListing(Command):
    @tool_required("unsquashfs")
    def cmdline(self):
        return ["unsquashfs", "-d", "", "-lls", self.path]


class SquashfsInvalidLineFormat(Exception):
    pass


class SquashfsMember(ArchiveMember):
    def is_directory(self):
        return False

    def is_symlink(self):
        return False

    def is_device(self):
        return False

    def is_socket_or_fifo(self):
        return False

    @property
    def path(self):
        # Use our extracted version and also avoid creating a temporary
        # directory per-file in ArchiveMember.path.
        return os.path.join(self.container._temp_dir, self._name)

    @property
    def name(self):
        # Don't include the leading "." in the output  (eg. "./etc/shadow")
        return self._name[1:]


class SquashfsRegularFile(SquashfsMember):
    # Example line:
    # -rw-r--r-- user/group   446 2015-06-24 14:49 squashfs-root/text
    LINE_RE = re.compile(
        r"^\S+\s+\S+\s+\S+\s+\S+\s+\S+\s+(?P<member_name>.*)$"
    )

    @staticmethod
    def parse(line):
        m = SquashfsRegularFile.LINE_RE.match(line)
        if not m:
            raise SquashfsInvalidLineFormat("invalid line format")
        return m.groupdict()

    def __init__(self, archive, member_name):
        SquashfsMember.__init__(self, archive, member_name)


class SquashfsDirectory(Directory, SquashfsMember):
    # Example line:
    # drwxr-xr-x user/group    51 2015-06-24 14:47 squashfs-root
    LINE_RE = re.compile(
        r"^\S+\s+\S+\s+\S+\s+\S+\s+\S+\s+(?P<member_name>.*)$"
    )

    @staticmethod
    def parse(line):
        m = SquashfsDirectory.LINE_RE.match(line)
        if not m:
            raise SquashfsInvalidLineFormat("invalid line format")
        return m.groupdict()

    def __init__(self, archive, member_name):
        SquashfsMember.__init__(self, archive, member_name or "/")

    def compare(self, other, source=None):
        return None

    def has_same_content_as(self, other):
        return False

    @property
    def path(self):
        raise NotImplementedError(
            "SquashfsDirectory is not meant to be extracted."
        )

    def is_directory(self):
        return True

    def get_member_names(self):
        raise ValueError("squashfs are compared as a whole.")  # noqa

    def get_member(self, member_name):
        raise ValueError("squashfs are compared as a whole.")  # noqa


class SquashfsSymlink(Symlink, SquashfsMember):
    # Example line:
    # lrwxrwxrwx user/group   6 2015-06-24 14:47 squashfs-root/link -> broken
    LINE_RE = re.compile(
        r"^\S+\s+\S+\s+\S+\s+\S+\s+\S+\s+(?P<member_name>.*)\s+->\s+(?P<destination>.*)$"
    )

    @staticmethod
    def parse(line):
        m = SquashfsSymlink.LINE_RE.match(line)
        if not m:
            raise SquashfsInvalidLineFormat("invalid line format")
        return m.groupdict()

    def __init__(self, archive, member_name, destination):
        SquashfsMember.__init__(self, archive, member_name)
        self._destination = destination

    def is_symlink(self):
        return True

    @property
    def symlink_destination(self):
        return self._destination


class SquashfsDevice(Device, SquashfsMember):
    # Example line:
    # crw-r--r-- root/root  1,  3 2015-06-24 14:47 squashfs-root/null
    LINE_RE = re.compile(
        r"^(?P<kind>c|b)\S+\s+\S+\s+(?P<major>\d+),\s*(?P<minor>\d+)\s+\S+\s+\S+\s+(?P<member_name>.*)$"
    )

    KIND_MAP = {"c": stat.S_IFCHR, "b": stat.S_IFBLK}

    @staticmethod
    def parse(line):
        m = SquashfsDevice.LINE_RE.match(line)
        if not m:
            raise SquashfsInvalidLineFormat("invalid line format")

        d = m.groupdict()
        try:
            d["mode"] = SquashfsDevice.KIND_MAP[d["kind"]]
            del d["kind"]
        except KeyError:
            raise SquashfsInvalidLineFormat(f"unknown device kind {d['kind']}")

        try:
            d["major"] = int(d["major"])
        except ValueError:
            raise SquashfsInvalidLineFormat(
                f"unable to parse major number {d['major']}"
            )

        try:
            d["minor"] = int(d["minor"])
        except ValueError:
            raise SquashfsInvalidLineFormat(
                f"unable to parse minor number {d['minor']}"
            )
        return d

    def __init__(self, archive, member_name, mode, major, minor):
        SquashfsMember.__init__(self, archive, member_name)
        self._mode = mode
        self._major = major
        self._minor = minor

    def get_device(self):
        return (self._mode, self._major, self._minor)

    def is_device(self):
        return True


class SquashfsFIFO(SocketOrFIFO, SquashfsMember):
    # Example line:
    # crw-r--r-- root/root  0 2021-08-18 13:37 run/initctl
    LINE_RE = re.compile(
        r"^(?P<kind>s|p)\S+\s+\S+\s+\S+\s+\S+\s+\S+\s+(?P<member_name>.*)$"
    )

    KIND_MAP = {"s": stat.S_IFSOCK, "p": stat.S_IFIFO}

    @staticmethod
    def parse(line):
        m = SquashfsFIFO.LINE_RE.match(line)
        if not m:
            raise SquashfsInvalidLineFormat("invalid line format")

        d = m.groupdict()
        try:
            d["mode"] = SquashfsFIFO.KIND_MAP[d["kind"]]
            del d["kind"]
        except KeyError:
            raise SquashfsInvalidLineFormat(
                f"unknown socket/FIFO kind {d['kind']}"
            )
        return d

    def __init__(self, archive, member_name, mode):
        SquashfsMember.__init__(self, archive, member_name)
        self._mode = mode

    def get_type(self):
        return stat.S_IFMT(self._mode)

    def is_socket_or_fifo(self):
        return True


class SquashfsContainer(Archive):
    auto_diff_metadata = False

    MEMBER_CLASS = {
        "d": SquashfsDirectory,
        "l": SquashfsSymlink,
        "c": SquashfsDevice,
        "b": SquashfsDevice,
        "p": SquashfsFIFO,
        "s": SquashfsFIFO,
        "-": SquashfsRegularFile,
    }

    def open_archive(self):
        return True

    def close_archive(self):
        pass

    def get_member(self, member_name):
        self.ensure_unpacked()
        cls, kwargs = self._members[member_name]
        return cls(self, member_name, **kwargs)

    def extract(self, member_name, destdir):
        # Ignore destdir argument and use our unpacked path
        self.ensure_unpacked()
        return member_name

    def get_member_names(self):
        self.ensure_unpacked()
        return self._members.keys()

    def ensure_unpacked(self):
        if hasattr(self, "_members"):
            return

        self._members = collections.OrderedDict()
        self._temp_dir_object = get_temporary_directory(suffix="squashfs")
        self._temp_dir = self._temp_dir_object.name

        logger.debug("Extracting %s to %s", self.source.path, self._temp_dir)

        cmd = (
            "unsquashfs",
            "-n",
            "-f",
            "-no",
            "-li",
            "-d",
            ".",
            os.path.abspath(self.source.path),
        )

        p = subprocess.Popen(
            cmd,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            cwd=self._temp_dir,
        )
        output, stderr = p.communicate()

        if p.returncode != 0:
            # unsquashfs(1) exits with 1 (with a suitable logging messages
            # that we can check for) if it could not extract, for example,
            # character devices that require superuser privileges. In this
            # case, don't treat this as a failure that requires reverting to
            # xxd(1), but do let the user know via a comment. After 4.5, this
            # return code can also be 2, to distinguish between fatal and
            # non-fatal errors (see upstream commit
            # 06fc061ef426672576075f2989f52acd873c0141)
            if (
                p.returncode in {1, 2}
                and b"because you're not superuser" in stderr
                and b"\n\ncreated " in output
            ):
                logger.debug("Ignoring unsquashfs return code")

                self.source.add_comment(
                    "Differences may be incomplete: {}".format(
                        stderr.decode("utf-8")
                    )
                )
            else:
                raise subprocess.CalledProcessError(
                    p.returncode, cmd, output, stderr
                )

        output = iter(output.decode("utf-8").rstrip("\n").split("\n"))

        # Skip headers
        for _ in iter(functools.partial(next, output), ""):
            pass

        for line in output:
            if not line:
                continue

            try:
                cls = self.MEMBER_CLASS[line[0]]
            except KeyError:
                logger.debug("Unknown squashfs entry: %s", line)
                continue

            try:
                kwargs = cls.parse(line)
            except SquashfsInvalidLineFormat:
                continue

            # Pop to avoid duplicating member name in the key and the value
            member_name = kwargs.pop("member_name")

            self._members[member_name] = (cls, kwargs)

        logger.debug(
            "Extracted %d entries from %s to %s",
            len(self._members),
            self.source.path,
            self._temp_dir,
        )


class SquashfsFile(File):
    DESCRIPTION = "SquashFS filesystems"
    CONTAINER_CLASSES = [SquashfsContainer]
    FILE_TYPE_RE = re.compile(r"^Squashfs filesystem\b")

    def compare_details(self, other, source=None):
        return [
            Difference.from_operation(
                SquashfsSuperblock, self.path, other.path
            ),
            Difference.from_operation(SquashfsListing, self.path, other.path),
        ]



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2016-2020, 2023 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import logging

from diffoscope.profiling import profile
from diffoscope.comparators.text import TextFile

from ...utils import format_class

from .. import ComparatorManager

logger = logging.getLogger(__name__)


def try_recognize(file, cls, recognizes):
    # Does this file class match?
    with profile("recognizes", file):
        # logger.debug("trying %s on %s", cls, file)
        if not recognizes(file):
            return False

    # Found a match; perform type magic
    logger.debug(
        "Using %s for %s",
        format_class(cls, strip="diffoscope.comparators."),
        file.name,
    )
    specialize_as(cls, file)

    return True


def specialize_as(cls, file):
    """
    Sometimes it is near-certain that files within a Container with a given
    extension (say) are of a known File type. We therefore do not need to run
    libmagic on these files. Using this facily within a Container can be an
    optimization, especially in cases where the Container contains hundreds of
    similar/small files. (This can be seen in the case of apktool and .smali
    files). In this case, this method can be used to essentially fix/force the
    type. Care should naturally be taken within Container implementations;
    such as checking the file extension and so forth.
    """

    new_cls = type(cls.__name__, (cls, type(file)), {})
    file.__class__ = new_cls
    return file


def specialize(file):
    # If we already know the class (ie. via `specialize_as`), then we do not
    # need to run `File.recognizes` at all.
    for cls in ComparatorManager().classes:
        if isinstance(file, cls):
            return file

    # Run the usual `File.recognizes` implementation.
    for cls in ComparatorManager().classes:
        if try_recognize(file, cls, cls.recognizes):
            return file

    # If there are no matches, run the fallback implementation.
    for cls in ComparatorManager().classes:
        if try_recognize(file, cls, cls.fallback_recognizes):
            logger.debug(
                "File recognized by fallback. Magic says: %s",
                file.magic_file_type,
            )
            return file

    logger.debug(
        "%s not identified by any comparator. Magic says: %s",
        file.name,
        file.magic_file_type,
    )

    if file.magic_mime_type == "text/plain":
        logger.debug(
            "However, %s is probably text; using %s",
            file.name,
            format_class(TextFile, strip="diffoscope.comparators."),
        )
        return specialize_as(TextFile, file)

    return file


def is_direct_instance(file, cls):
    # is_direct_instance(<GzipFile>, GzipFile) == True, but
    # is_direct_instance(<IpkFile>, GzipFile) == False
    if not isinstance(file, cls):
        return False
    for c in ComparatorManager().classes:
        if c is not cls and isinstance(file, c):
            return False
    return True



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2017-2020 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import fnmatch
import logging
import re

from diffoscope.config import Config

logger = logging.getLogger(__name__)


def operation_excluded(operation):
    for y in Config().exclude_commands:
        if re.search(y, operation):
            logger.debug(
                "Excluding command '%s' as it matches pattern '%s'",
                operation,
                y,
            )
            return True
    return False


def filter_excludes(filenames):
    for x in filenames:
        for y in Config().excludes:
            if fnmatch.fnmatchcase(x, y):
                logger.debug("Excluding %s as it matches pattern '%s'", x, y)
                break
        else:
            yield x


def any_excluded(*filenames):
    for x in filenames:
        for y in Config().excludes:
            if fnmatch.fnmatchcase(x, y):
                return True
    return False



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2015-2020, 2023 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import re
import base64
import logging
import subprocess

from diffoscope.config import Config
from diffoscope.tools import tool_required
from diffoscope.tempfiles import get_named_temporary_file
from diffoscope.difference import Difference, VisualDifference

from .utils.file import File
from .utils.command import Command, our_check_output

re_ansi_escapes = re.compile(r"\x1b[^m]*m")

logger = logging.getLogger(__name__)


class Img2Txt(Command):
    @tool_required("img2txt")
    def cmdline(self):
        return ["img2txt", "--width", "60", "--format", "utf8", self.path]

    def filter(self, line):
        # Strip ANSI escapes
        return re_ansi_escapes.sub("", line.decode("utf-8")).encode("utf-8")


class Identify(Command):
    ATTRIBUTES = (
        "Image format: %m",
        "Height: %[height]",
        "Width: %[width]",
        "Orientation: %[orientation]",
        "Compression type: %[compression]",
        "Compression quality: %[quality]",
        "Colorspace: %[colorspace]",
        "Channels: %[channels]",
        "Depth: %[depth]",
        "Interlace mode: %[interlace]",
        "Rendering intent: %[rendering-intent]",
        "X resolution: %[resolution.x]",
        "Y resolution: %[resolution.y]",
        "Resolution units: %[units]",
        "Transparency channel enabled: %A",
        "Gamma: %[gamma]",
        "Number of unique colors: %[colors]",
        "Comment: %c",
        "EXIF data: %[EXIF:*]",
    )

    @tool_required("identify")
    def cmdline(self):
        return ["identify", "-format", "\n".join(self.ATTRIBUTES), self.path]


@tool_required("compare")
def pixel_difference(image1_path, image2_path):
    compared_filename = get_named_temporary_file(suffix=".png").name

    try:
        our_check_output(
            (
                "compare",
                image1_path,
                image2_path,
                "-compose",
                "src",
                compared_filename,
            )
        )
    except subprocess.CalledProcessError as e:
        # ImageMagick's `compare` will return 1 if images are different
        if e.returncode == 1:
            pass

    with open(compared_filename, "rb") as f:
        content = base64.b64encode(f.read()).decode("utf8")

    return VisualDifference("image/png;base64", content, "Pixel difference")


@tool_required("convert")
def flicker_difference(image1_path, image2_path):
    compared_filename = get_named_temporary_file(suffix=".gif").name

    our_check_output(
        (
            "convert",
            "-delay",
            "50",
            image1_path,
            image2_path,
            "-loop",
            "0",
            "-compose",
            "difference",
            compared_filename,
        )
    )

    with open(compared_filename, "rb") as f:
        content = base64.b64encode(f.read()).decode("utf8")

    return VisualDifference("image/gif;base64", content, "Flicker difference")


@tool_required("identify")
def get_image_size(image_path):
    return our_check_output(("identify", "-format", "%[h]x%[w]", image_path))


def same_size(image1, image2):
    try:
        return get_image_size(image1.path) == get_image_size(image2.path)
    except subprocess.CalledProcessError:  # noqa
        return False


class JPEGImageFile(File):
    DESCRIPTION = "JPEG images"
    FILE_TYPE_RE = re.compile(r"\bJPEG image data\b")

    def compare_details(self, other, source=None):
        content_diff = Difference.from_operation(
            Img2Txt, self.path, other.path, source="Image content"
        )
        if (
            content_diff is not None
            and Config().compute_visual_diffs
            and same_size(self, other)
        ):
            try:
                logger.debug(
                    "Generating visual difference for %s and %s",
                    self.path,
                    other.path,
                )
                content_diff.add_visuals(
                    [
                        pixel_difference(self.path, other.path),
                        flicker_difference(self.path, other.path),
                    ]
                )
            except subprocess.CalledProcessError:  # noqa
                pass
        return [
            content_diff,
            Difference.from_operation(
                Identify, self.path, other.path, source="Image metadata"
            ),
        ]


class ICOImageFile(File):
    DESCRIPTION = "Microsoft Windows icon files"
    FILE_TYPE_RE = re.compile(r"\bMS Windows icon resource\b")

    def compare_details(self, other, source=None):
        differences = []

        # img2txt does not support .ico files directly so convert to .PNG.
        try:
            png_a, png_b = [ICOImageFile.convert(x) for x in (self, other)]
        except subprocess.CalledProcessError:  # noqa
            pass
        else:
            content_diff = Difference.from_operation(
                Img2Txt, png_a, png_b, source="Image content"
            )
            if (
                content_diff is not None
                and Config().compute_visual_diffs
                and same_size(self, other)
            ):
                if get_image_size(self.path) == get_image_size(other.path):
                    logger.debug(
                        "Generating visual difference for %s and %s",
                        self.path,
                        other.path,
                    )
                    content_diff.add_visuals(
                        [
                            pixel_difference(self.path, other.path),
                            flicker_difference(self.path, other.path),
                        ]
                    )
            differences.append(content_diff)

        differences.append(
            Difference.from_operation(
                Identify, self.path, other.path, source="Image metadata"
            )
        )

        return differences

    @staticmethod
    @tool_required("convert")
    def convert(file):
        result = get_named_temporary_file(suffix=".png").name

        our_check_output(("convert", file.path, result))

        return result



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2020-2021 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import os
import time
import logging

logger = logging.getLogger(__name__)


def normalize_environment():
    """
    Normalise locale, etc. so external tools give us stable and properly
    encoded output.
    """

    logger.debug(
        "Normalising locale, timezone, etc. PATH is %s",
        os.environ.get("PATH", "(none)"),
    )

    for x in ("LANGUAGE", "LC_ALL"):
        os.environ.pop(x, None)

    for x in (
        "LANG",
        "LC_NUMERIC",
        "LC_TIME",
        "LC_COLLATE",
        "LC_MONETARY",
        "LC_MESSAGES",
        "LC_PAPER",
        "LC_NAME",
        "LC_ADDRESS",
        "LC_TELEPHONE",
        "LC_MEASUREMENT",
        "LC_IDENTIFICATION",
    ):
        os.environ[x] = "C"

    os.environ["TZ"] = "UTC"
    os.environ["LC_CTYPE"] = "C.UTF-8"

    time.tzset()



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2017-2022 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import sys
import codecs
import collections
import contextlib
import string
import _string


def round_sigfig(num, s):
    # https://stackoverflow.com/questions/3410976/how-to-round-a-number-to-significant-figures-in-python
    # This was too painful :/
    x = float(("%%.%sg" % s) % num)
    return x if abs(x) < (10 ** (s - 1)) else int(x)


def sizeof_fmt(num, suffix="B", sigfig=3):
    # https://stackoverflow.com/questions/1094841/reusable-library-to-get-human-readable-version-of-file-size
    # A more powerful version is python3-hurry.filesize but that's an extra dependency
    for unit in ["", "K", "M", "G", "T", "P", "E", "Z"]:
        if abs(num) < 1024.0:
            break
        num /= 1024.0
    else:
        unit = "Y"
    return "{} {}{}".format(round_sigfig(num, sigfig), unit, suffix)


class Presenter:
    supports_visual_diffs = False

    def __init__(self):
        self.depth = 0

    @classmethod
    def run(cls, data, difference, parsed_args):
        with make_printer(data["target"]) as fn:
            cls(fn).start(difference)

    def start(self, difference):
        self.visit(difference)

    def visit(self, difference):
        self.visit_difference(difference)

        self.depth += 1

        for x in difference.details:
            self.visit(x)

        self.depth -= 1

    def visit_difference(self, difference):
        raise NotImplementedError()

    @classmethod
    def indent(cls, val, prefix):
        # As an optimisation, output as much as possible in one go to avoid
        # unnecessary splitting, interpolating, etc.
        #
        # We don't use textwrap.indent as that unnecessarily calls
        # str.splitlines, etc.
        return prefix + val.rstrip().replace("\n", "\n{}".format(prefix))


class PrintLimitReached(Exception):
    pass


class DiffBlockLimitReached(Exception):
    pass


@contextlib.contextmanager
def make_printer(path):
    output = sys.stdout

    if path != "-":
        output = codecs.open(path, "w", encoding="utf-8")

    def fn(*args, **kwargs):
        kwargs["file"] = output
        print(*args, **kwargs)

    fn.output = output

    yield fn

    if path != "-":
        output.close()


def create_limited_print_func(print_func, max_page_size):
    count = [0]

    def fn(val, force=False):
        print_func(val)

        if force or max_page_size == 0:
            return

        count[0] += len(val)
        if count[0] >= max_page_size:
            raise PrintLimitReached()

    return fn


class PartialFormatter(string.Formatter):
    @staticmethod
    def escape(x):
        return x.replace("}", "}}").replace("{", "{{")

    def get_value(self, key, args, kwargs):
        return args[key] if isinstance(key, int) else args[int(key)]

    def arg_of_field_name(self, field_name, args):
        x = int(_string.formatter_field_name_split(field_name)[0])
        return x if x >= 0 else len(args) + x

    def parse(self, *args, **kwargs):
        # Preserve {{ and }} escapes when formatting
        return map(
            lambda x: (self.escape(x[0]),) + x[1:],
            super().parse(*args, **kwargs),
        )

    parse_no_escape = string.Formatter.parse


class FormatPlaceholder:
    def __init__(self, ident):
        self.ident = str(ident)

    def __repr__(self):
        return f"{self.__class__.__name__}({self.ident!r})"

    def __format__(self, spec):
        result = self.ident
        if spec:
            result += ":" + spec
        return f"{{{result}}}"

    def __getitem__(self, key):
        return FormatPlaceholder(self.ident + "[" + str(key) + "]")

    def __getattr__(self, attr):
        return FormatPlaceholder(self.ident + "." + str(attr))


class PartialString:
    r"""A format string where the "holes" are indexed by arbitrary python
    objects instead of string names or integer indexes. This is useful when you
    need to compose these objects together, but don't want users of the partial
    string to have to deal with artificial "indexes" for the holes.

    For example:

    >>> a, b = object(), object()
    >>> tmpl = PartialString("{0} {1}", a, b)
    >>> tmpl
    PartialString('{0} {1}', <object object at ...>, <object object at ...>)
    >>> tmpl.holes == (a, b)
    True
    >>> tmpl.format({a: "Hello,", b: "World!"})
    'Hello, World!'

    You can partially fill up the holes:

    >>> tmpl.pformat({a: "Hello,"}) == PartialString('Hello, {0}', b)
    True
    >>> tmpl.pformat({b: "World!"}) == PartialString('{0} World!', a)
    True

    You can estimate the size of the filled-up string:

    >>> tmpl.base_len, tmpl.num_holes
    (1, 2)
    >>> tmpl.size(hole_size=33)
    67

    You can also partially fill up the holes using more PartialStrings,
    even recursively:

    >>> tmpl.pformat({a: PartialString('{0}', b)}) == PartialString('{0} {0}', b)
    True
    >>> tmpl.pformat({a: tmpl}) == PartialString('{0} {1} {1}', a, b)
    True
    >>> tmpl.pformat({b: tmpl}) == PartialString('{0} {0} {1}', a, b)
    True

    Finally, the holes have to match what's in the format string:

    >>> tmpl = PartialString("{0} {1} {2}", a, b)
    Traceback (most recent call last):
    ...
    IndexError: tuple index out of range

    If you don't have specific objects to index the holes with, and don't want
    to create artificial indexes as in the above examples, here is a slightly
    simpler way of doing it:

    >>> tmpl = PartialString.numl("{0} {1} {2}", 2, object())
    >>> tmpl.holes
    (0, 1, <object ...>)
    >>> tmpl.pformatl("(first hole)", "(second hole)", "(object hole)")
    PartialString('(first hole) (second hole) (object hole)',)

    CORNER CASES:

    1. If you need to include a literal '{' or '}' in the resulting formatted
    string, you need to give them as "{{" or "}}" respectively in the fmtstr
    parameter of PartialString.__init__. PartialString.escape() might help to
    make this a bit easier:

    >>> tmpl = PartialString.numl("find {0} -name {1} " +
    ...            PartialString.escape("-exec ls -la {} \;"), 2)
    >>> tmpl
    PartialString('find {0} -name {1} -exec ls -la {{}} \\;', ...)
    >>> tmpl.size(), tmpl.size(4)
    (33, 39)

    When using pformat, any string arguments will be escaped automatically. You
    can take advantage of this to simplify the above example:

    >>> tmpl2 = PartialString.numl("find {0} -name {1} -exec ls -la {2} \;", 3)
    >>> tmpl2 = tmpl2.pformat({2: "{}"})
    >>> tmpl2 == tmpl
    True

    As long as you only use pformat, any "{{" "}}" literals will remain escaped
    in the resulting PartialString. They only become unescaped after going
    through a full format.

    >>> tmpl.pformatl("my{}path", "my{}file")
    PartialString('find my{{}}path -name my{{}}file -exec ls -la {{}} \\;',)
    >>> tmpl.formatl("my{}path", "my{}file")
    'find my{}path -name my{}file -exec ls -la {} \\;'

    CAVEATS:

    1. Filling up holes using other PartialStrings, does not play very nicely
    with format specifiers. For example:

    >>> tmpl = PartialString("{0:20} {1.child}", a, b)
    >>> tmpl.pformat({a: tmpl})
    PartialString('{0:20} {1.child}     {1.child}', <object ...>, <object ...>)
    >>> tmpl.pformat({b: tmpl})
    Traceback (most recent call last):
    ...
    AttributeError: ... has no attribute 'child'

    So you probably want to avoid such usages. The exact behaviour of these
    might change in the future, too.
    """

    formatter = PartialFormatter()
    escape = staticmethod(PartialFormatter.escape)

    def __init__(self, fmtstr="", *holes):
        # Ensure the format string is valid, and figure out some basic stats
        fmt = self.formatter
        # use parse_no_escape so lengths are preserved
        pieces = [(len(l), f) for l, f, _, _ in fmt.parse_no_escape(fmtstr)]
        used_args = set(
            fmt.arg_of_field_name(f, holes) for _, f in pieces if f is not None
        )
        self.num_holes = sum(1 for _, f in pieces if f is not None)
        self.base_len = sum(l for l, _ in pieces)

        # Remove unused and duplicates in the holes objects
        seen = collections.OrderedDict()
        mapping = tuple(
            (
                FormatPlaceholder(seen.setdefault(k, len(seen)))
                if i in used_args
                else None
            )
            for i, k in enumerate(holes)
        )
        self._fmtstr = fmt.vformat(fmtstr, mapping, None)
        self.holes = tuple(seen.keys())

    def __eq__(self, other):
        return (
            self is other
            or isinstance(other, PartialString)
            and other._fmtstr == self._fmtstr
            and other.holes == self.holes
        )

    def __repr__(self):
        return "%s%r" % (self.__class__.__name__, (self._fmtstr,) + self.holes)

    def _format(self, *mapping):
        # format a string but preserve {{ and }} escapes
        return self.formatter.vformat(self._fmtstr, mapping, None)

    def _offset_fmtstr(self, offset):
        return self._format(
            *(FormatPlaceholder(i + offset) for i in range(len(self.holes)))
        )

    def _pformat(self, mapping, escapestr):
        new_holes = []
        real_mapping = []
        for i, k in enumerate(self.holes):
            if k in mapping:
                v = mapping[k]
                if isinstance(v, PartialString):
                    out = v._offset_fmtstr(len(new_holes))
                    new_holes.extend(v.holes)
                elif isinstance(v, str) and escapestr:
                    out = PartialString.escape(v)
                else:
                    out = v
            else:
                out = FormatPlaceholder(len(new_holes))
                new_holes.append(k)
            real_mapping.append(out)
        return real_mapping, new_holes

    def size(self, hole_size=1):
        return self.base_len + hole_size * self.num_holes

    def pformat(self, mapping=None):
        """Partially apply a mapping, returning a new PartialString."""
        if mapping is None:
            mapping = {}
        real_mapping, new_holes = self._pformat(mapping, True)
        return self.__class__(self._format(*real_mapping), *new_holes)

    def pformatl(self, *args):
        """Partially apply a list, implicitly mapped from self.holes."""
        return self.pformat(dict(zip(self.holes, args)))

    def format(self, mapping=None):
        """Fully apply a mapping, returning a string."""
        if mapping is None:
            mapping = {}
        real_mapping, new_holes = self._pformat(mapping, False)
        if new_holes:
            raise ValueError(f"not all holes filled: {new_holes!r}")
        return self._fmtstr.format(*real_mapping)

    def formatl(self, *args):
        """Fully apply a list, implicitly mapped from self.holes."""
        return self.format(dict(zip(self.holes, args)))

    @classmethod
    def of(cls, obj):
        """Create a partial string for a single object.

        >>> e = PartialString.of(None)
        >>> e.pformat({None: e}) == e
        True
        """
        return cls("{0}", obj)

    @classmethod
    def numl(cls, fmtstr="", nargs=0, *holes):
        """Create a partial string with some implicit numeric holes.

        Useful in conjunction with PartialString.pformatl.

        >>> PartialString.numl("{0}{1}{2}{3}", 3, "last object")
        PartialString('{0}{1}{2}{3}', 0, 1, 2, 'last object')
        >>> _.pformatl(40, 41, 42, "final")
        PartialString('404142final',)
        """
        return cls(fmtstr, *range(nargs), *holes)

    @classmethod
    def cont(cls):
        r"""Create a new empty partial string with a continuation token.

        Construct a larger partial string from smaller pieces, without having
        to keep explicit track of a global index in between pieces. Instead,
        you can use per-piece local indexes, plus the special index {-1} to
        refer to where the next piece will go - or omit it to end the sequence.

        >>> t, cont = PartialString.cont()
        >>> t = cont(t, "x: {0}\ny: {1}\n{-1}", object(), object())
        >>> t = cont(t, "z: {0}\n{-1}", object())
        >>> t = cont(t, "")
        >>> key = t.holes
        >>> t.format({key[0]: "line1", key[1]: "line2", key[2]: "line3"})
        'x: line1\ny: line2\nz: line3\n'
        >>> t.size(hole_size=5)
        27
        """

        def cont(t, fmtstr, *holes):
            if isinstance(fmtstr, cls):
                return t.pformat({cont: fmtstr})
            return t.pformat({cont: cls(fmtstr, *(holes + (cont,)))})

        return cls("{0}", cont), cont

    def frame(self, header, footer):
        frame = self.__class__(
            self.escape(header) + "{0}" + self.escape(footer), None
        )
        return frame.pformat({None: self})


if __name__ == "__main__":
    import doctest

    doctest.testmod(optionflags=doctest.ELLIPSIS)
    a, b = object(), object()
    tmpl = PartialString("{0} {1}", a, b)
    t, cont = PartialString.cont()



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2015 Reiner Herrmann <reiner@reiner-h.de>
# Copyright © 2016-2020 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import re
import os.path
import logging
import subprocess

from diffoscope.difference import Difference
from diffoscope.tools import tool_required, tool_check_installed

from .utils.file import File
from .utils.archive import Archive
from .utils.command import Command

logger = logging.getLogger(__name__)


class DexContainer(Archive):
    @property
    def path(self):
        return self._path

    def open_archive(self):
        return self

    def close_archive(self):
        pass

    def get_member_names(self):
        return [self.get_compressed_content_name(".dex") + ".jar"]

    @tool_required("enjarify")
    def extract(self, member_name, dest_dir):
        dest_path = os.path.join(dest_dir, member_name)
        logger.debug("dex extracting to %s", dest_path)
        subprocess.check_call(
            ["enjarify", "-o", dest_path, self.source.path],
            stderr=None,
            stdout=subprocess.PIPE,
        )
        return dest_path


class DexFile(File):
    DESCRIPTION = "Dalvik .dex files"
    FILE_TYPE_HEADER_PREFIX = b"dex\n"
    FILE_TYPE_RE = re.compile(r"^Dalvik dex file\b")
    FILE_EXTENSION_SUFFIX = {".dex"}
    CONTAINER_CLASSES = [DexContainer]

    def compare_details(self, other, source=None):
        if tool_check_installed("dexdump"):
            dexdump = Difference.from_operation(Dexdump, self.path, other.path)
            if dexdump is not None:
                dexdump.add_comment(
                    "Ignoring differences in offsets to keep diff size reasonable."
                )
                return [dexdump]
        else:
            self.add_comment(
                "'dexdump' command not installed; cannot compare format-specific details."
            )
        return []


class Dexdump(Command):
    @tool_required("dexdump")
    def cmdline(self):
        # -a : display annotations
        # -d : disassemble code sections
        # -f : display summary information from file header
        # -h : display file header details
        return ["dexdump", "-a", "-d", "-f", "-h", self.path]

    def filter(self, line):
        # Processing 'classes.dex'...
        if line.startswith(b"Processing "):
            return b""

        # Opened 'classes.dex', DEX version '037'
        if line.startswith(b"Opened "):
            return line.split(b", ", 1)[-1]

        # ignore differences in offsets
        offsets = (b"annotations", b"class_data", b"interfaces")
        if any(line.startswith(o + b"_off") for o in offsets):
            return b""

        return line
